/*
 * spool.c: Handle disk-cache for videotext-pages
 *
 * Copyright (c) 1994-96 Martin Buck  <martin-2.buck@student.uni-ulm.de>
 * Read COPYING for more information
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/vtx.h>
#include <xview/xview.h>
#include <xview/notify.h>
#include "safe_malloc.h"
#include "tilde.h"
#include "misc.h"
#include "config.h"
#include "hotlist.h"
#include "vtxqueue.h"
#include "cache.h"
#include "vtxtools.h"
#include "xinit.h"
#include "spool.h"


#define NICE_VAL 5



static int
write_spooldir(const char *basedir, int private) {
  char *ch, *dirname, *fname_offs, station_name[STATIONNAME_MAXLEN + 1], buffer[VTX_PAGESIZE];
  int oldumask;
  cache_loop_t last;
  vtx_pageinfo_t info;
  FILE *file;
  
  dirname = sstrdup(basedir);
  dirname = sstrapp(dirname, "/");
  strcpy(station_name, hotlist_get_name(hotlist_get_current()));
  for (ch = station_name; *ch; ch++) {
    if (*ch == '/') {
      *ch = '_';
    }
  }
  dirname = sstrapp(dirname, station_name);
  oldumask = umask(0);
  if (private) {
    mkdir(dirname, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
  } else {
    mkdir(dirname, S_IRWXU | S_IRWXG | S_IRWXO);
  }
  umask(0022);
  if (access(dirname, R_OK | W_OK | X_OK) < 0) {
    umask(oldumask);
    free(dirname);
    return -1;
  }
  dirname = sstrapp(dirname, "/??????????");
  fname_offs = dirname + strlen(dirname) - 10;
  last.page = -1;
  while (cache_loop(&last, buffer, &info)) {
    if (hotlist_check_spool(info.pagenum)) {
      sprintf(fname_offs, "%X_%X.vtx", info.pagenum, info.minute);
      unlink(dirname);
      if ((file = fopen(dirname, "w"))) {
        save_vtx(file, buffer, &info, FALSE);
        fclose(file);
      }
    }
  }
  umask(oldumask);
  free(dirname);
  return 0;
}


void
write_cache(void) {
  int fd, maxopen;
  pid_t pid;
  const char *privdir;
  
  if (!hotlist_get_current() || config_get_spool_only())
    return;

  if ((pid = fork()) < 0) {	/* ignore failure because writing the cache to disk is not vital */
    return;
  } else if (pid == 0) {	/* this is the new process */
    if ((maxopen = sysconf(_SC_OPEN_MAX)) < 0) {
      maxopen = 256;		/* hopefully enough */
    }
    for (fd = 0; fd < maxopen; fd++) {
      close(fd);
    }
    nice(NICE_VAL);		/* No need to hurry */

    if (hotlist_get_spoolflag(TRUE, FALSE) && strcmp(privdir = config_get_spooldir(), "") &&
        (privdir = tilde_expand(privdir))) {
      write_spooldir(privdir, TRUE);
    }
    if (hotlist_get_spoolflag(FALSE, FALSE)) {
      write_spooldir(VTX_SPOOLDIR, FALSE);
    }
    _exit(0);
  }
  notify_set_wait3_func(frame, notify_default_wait3, pid);	/* reap child when finished */
}


static int
read_spool(const char *basedir, int page, int subpage, byte_t *buffer, vtx_pageinfo_t *info) {
  FILE *file;
  struct stat st;
  char *fname, *fname_offs, *dirname, station_name[STATIONNAME_MAXLEN + 1], *ch;
  int retval = FALSE;
  
  dirname = sstrdup(basedir);
  dirname = sstrapp(dirname, "/");
  strcpy(station_name, hotlist_get_name(hotlist_get_current()));
  for (ch = station_name; *ch; ch++) {
    if (*ch == '/') {
      *ch = '_';
    }
  }
  dirname = sstrapp(dirname, station_name);
  if (access(dirname, R_OK | X_OK) < 0) {
    free(dirname);
    return FALSE;
  }
  fname = NULL;
  if (subpage == -1) {					/* Find the first subpage */
    DIR *dir;
    struct dirent *de;
    char fntempl[11];
    int f_sub;
    
    if ((dir = opendir(dirname))) {
      sprintf(fntempl, "%X_%%d.vtx", page);
      while ((de = readdir(dir))) {
        if (sscanf(de->d_name, fntempl, &f_sub) == 1 && (subpage < 0 || f_sub < subpage)) {
          subpage = f_sub;
          if (subpage == 0 || subpage == 1)
            break;
        }
      }
      closedir(dir);
    }
  }
  if (subpage < 0) {
    free(dirname);
    return FALSE;
  }
  fname = sstrdup(dirname);
  fname = sstrapp(fname, "/??????????");
  fname_offs = fname + strlen(fname) - 10;
  sprintf(fname_offs, "%X_%X.vtx", page, subpage);
  if ((file = fopen(fname, "r"))) {
    if (!fstat(fileno(file), &st) &&
        (st.st_mtime / 3600 + hotlist_get_expire() >= time(NULL) / 3600 ||
        config_get_spool_only()) && load_vtx(file, vtx_buffer, &vtxwin.page->info, NULL) >= 0) {
      retval = TRUE;
    }
    fclose(file);
  }
  free(fname);
  free(dirname);
  return retval;
}


int
query_spool(int page, int subpage, byte_t *buffer, vtx_pageinfo_t *info) {
  int so;
  const char *privdir;
  
  so = config_get_spool_only();
  if (!vtx_chkpgnum(page, FALSE) || !hotlist_check_spool(page))
    return FALSE;
  if (((hotlist_get_spoolflag(TRUE, TRUE) || so) && strcmp(privdir = config_get_spooldir(), "") &&
      (privdir = tilde_expand(privdir)) && read_spool(privdir, page, subpage, buffer, info)) ||
      ((hotlist_get_spoolflag(FALSE, TRUE) || so) &&
      read_spool(VTX_SPOOLDIR, page, subpage, buffer, info))) {
    return TRUE;
  }
  return FALSE;
}
